@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -22,6 +22,8 @@
 # TODO: Use overlap (rather than simple pointer location) to determine
 # drag and drop.
 
+from __future__ import print_function
+
 import renpy.display
 from renpy.display.render import render, Render, redraw
 from renpy.display.core import absolute
@@ -53,7 +55,7 @@ def default_drag_joined(drag):
 class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
     """
     :doc: drag_drop class
-    :args: (d=None, drag_name=None, draggable=True, droppable=True, drag_raise=True, dragged=None, dropped=None, drag_handle=(0.0, 0.0, 1.0, 1.0), drag_joined=..., clicked=None, hovered=None, unhovered=None, **properties)
+    :args: (d=None, drag_name=None, draggable=True, droppable=True, drag_raise=True, dragged=None, dropped=None, drag_handle=(0.0, 0.0, 1.0, 1.0), drag_joined=..., clicked=None, hovered=None, unhovered=None, mouse_drop=False, **properties)
 
     A displayable that represents an object that can be dragged around
     its enclosing area. A Drag can also represent an area that
@@ -105,6 +107,12 @@ class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
         If true, this Drag is raised to the top when it is dragged. If
         it is joined to other Drags, all joined drags are raised.
 
+    `activated`
+        A callback (or list of callbacks) that is called when the mouse
+        is pressed down on the drag. It is called with one argument, a
+        a list of Drags that are being dragged. The return value of this
+        callback is ignored.
+
     `dragged`
         A callback (or list of callbacks) that is called when the Drag
         has been dragged. It is called with two arguments. The first is
@@ -149,6 +157,11 @@ class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
         size, as the drags can leave the screen entirely, with no
         way to get them back on the screen.
 
+    `mouse_drop`
+        If true, the drag is dropped on the first droppable under the cursor.
+        If false, the default, the drag is dropped onto the droppable with
+        the largest degree of overlap.
+
     Except for `d`, all of the parameters are available as fields (with
     the same name) on the Drag object. In addition, after the drag has
     been rendered, the following fields become available:
@@ -165,6 +178,7 @@ class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
     drag_group = None
     old_position = None
     drag_offscreen = False
+    activated = None
 
     def __init__(self,
                  d=None,
@@ -181,6 +195,8 @@ class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
                  unhovered=None,
                  replaces=None,
                  drag_offscreen=False,
+                 mouse_drop=False,
+                 activated=None,
                  style="drag",
                  **properties):
 
@@ -197,7 +213,12 @@ class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
         self.clicked = clicked
         self.hovered = hovered
         self.unhovered = unhovered
+        self.activated = activated
         self.drag_offscreen = drag_offscreen
+        # if mouse_drop_check is True (default False), the drop will not
+        #  use default major overlap between droppables but instead
+        #  will use mouse coordinates to select droppable
+        self.mouse_drop = mouse_drop
 
         # We're focusable if we can be dragged.
         self.focusable = draggable
@@ -272,6 +293,7 @@ class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
             self.old_position = replaces.old_position
             self.drag_moved = replaces.drag_moved
             self.last_drop = replaces.last_drop
+            self.mouse_drop = replaces.mouse_drop
 
         if d is not None:
             self.add(d)
@@ -345,6 +367,16 @@ class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
         if self.drag_group is not None:
             self.drag_group.raise_children([ self ])
 
+    def bottom(self):
+        """
+        :doc: drag_drop method
+
+        Lowers this displayable to the bottom of its drag_group.
+        """
+
+        if self.drag_group is not None:
+            self.drag_group.lower_children([ self ])
+
     def visit(self):
         return [ self.child ]
 
@@ -508,6 +540,8 @@ class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
 
             renpy.display.focus.set_grab(self)
 
+            run(joined[0].activated, joined)
+
             self.grab_x = x
             self.grab_y = y
 
@@ -581,7 +615,10 @@ class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
             handled = False
 
         if (self.drag_group is not None) and self.drag_moved:
-            drop = self.drag_group.get_best_drop(joined)
+            if self.mouse_drop:
+                drop = self.drag_group.get_drop_at(joined, par_x, par_y)
+            else:
+                drop = self.drag_group.get_best_drop(joined)
         else:
             drop = None
 
@@ -752,6 +789,20 @@ class DragGroup(renpy.display.layout.MultiBox):
         self.children = self._list_type(children)
         self.offsets = self._list_type(offsets)
 
+    def lower_children(self, l):
+        """
+        Lowers the children in `l` to the bottom of this drag group, with
+        the one at the bottom being the lowest.
+        """
+
+        self.children.reverse()
+        self.offsets.reverse()
+
+        self.raise_children(l)
+
+        self.children.reverse()
+        self.offsets.reverse()
+
     def get_best_drop(self, joined):
         """
         Returns the droppable that the members of joined overlap the most.
@@ -789,6 +840,26 @@ class DragGroup(renpy.display.layout.MultiBox):
         else:
             return rv
 
+    def get_drop_at(self, joined, x, y):
+        """
+        Returns the droppable that is exactly at x, y.
+        """
+
+        joined_set = set(joined)
+        for c in self.children:
+            if c in joined_set:
+                continue
+
+            if not c.droppable:
+                continue
+
+            if c.x is None:
+                continue
+
+            if (x >= c.x and y >= c.y and
+                    x < (c.x+c.w) and y < (c.y+c.h)):
+                return c
+
     def get_children(self):
         """
         Returns a list of Drags that are the children of
